"""
Data structures module for SmartyStreets API.

These structures simply wrap Python built in data structures that match the API's JSON responses,
including some convenience methods for simple access.
"""


class Address(dict):
    """
    Class for handling a single address response
    """

    @property
    def location(self):
        """
        Returns the geolocation as a lat/lng pair
        """
        try:
            lat, lng = self['metadata']['latitude'], self['metadata']['longitude']
        except KeyError:
            return None
        if not lat or not lng:
            return None
        return lat, lng

    @property
    def confirmed(self):
        """
        Returns a boolean whether this address is DPV confirmed

        The property does not specify *how* or what extent.

        """
        valid = ['Y', 'S', 'D']
        match_code = self.get('analysis', {}).get('dpv_match_code', '')
        return match_code in valid

    @property
    def id(self):
        """
        Returns the input id
        """
        try:
            return self['input_id']
        except KeyError:
            return None

    @property
    def index(self):
        """
        Returns the input_index
        """
        try:
            return self['input_index']
        except KeyError:
            return None


class AddressCollection(list):
    """
    Class for handling multiple responses.
    """
    id_lookup = {}  # For user supplied input_id
    index_lookup = {}  # For SmartyStreets input_index

    def __init__(self, results):
        """
        Constructor for an AddressCollection

        :param addresses: a list of dictionaries providing address information
        :return:
        """
        addresses = []
        for index, result in enumerate(results):
            address = Address(result)
            addresses.append(address)
            self.index_lookup[address.index] = index
            if address.id:
                self.id_lookup[address.id] = index
        super(AddressCollection, self).__init__(addresses)

    def get(self, key):
        """
        Returns an address by user controlled input ID

        :param key: an input_id used to tag a lookup address
        :return: a matching Address
        """
        try:
            return self[self.id_lookup.get(key)]
        except TypeError:
            raise KeyError

    def get_index(self, key):
        """
        Returns an address by input index, a value that matches the list index of the provided
        lookup value, not necessarily the result.

        :param key: an input_index matching the index of the provided address
        :return: a matching Address
        """
        try:
            return self[self.index_lookup.get(key)]
        except TypeError:
            raise KeyError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_smartystreets
----------------------------------

Tests for `smartystreets` module.
"""

import unittest

from smartystreets.data import Address, AddressCollection


class TestAddress(unittest.TestCase):

    def test_empty(self):
        addr = Address([])
        self.assertIsNone(addr.location)

    def test_location(self):
        addr = Address({'metadata': {'latitude': 12, 'longitude': 13}})
        self.assertEqual(addr.location, (12, 13))

    def test_id(self):
        addr = Address({'input_id': 'test1'})
        self.assertEqual(addr.id, 'test1')

    def test_index(self):
        addr = Address({'input_id': 'test1', 'input_index': 4})
        self.assertEqual(addr.index, 4)


class TestCollection(unittest.TestCase):

    def test_get_by_id(self):
        """Ensure an address can be found by input_id"""
        collection = AddressCollection([
            {'input_index': 3, 'input_id': 'A'},
            {'input_index': 5, 'input_id': 'h'},
            {'input_index': 8, 'input_id': 'X'},
        ])
        self.assertEqual(collection.get('A').index, 3)
        self.assertEqual(collection.get('h').index, 5)
        self.assertEqual(collection.get('X').index, 8)

        self.assertRaises(KeyError, collection.get_index, 'kj')

    def test_get_by_index(self):
        """Ensure an address can be found by input index"""
        collection = AddressCollection([
            {'input_index': 3, 'input_id': 'A'},
            {'input_index': 5, 'input_id': 'h'},
            {'input_index': 8, 'input_id': 'X'},
        ])
        self.assertEqual(collection.get_index(3).id, 'A')
        self.assertEqual(collection.get_index(5).id, 'h')
        self.assertEqual(collection.get_index(8).id, 'X')

        self.assertRaises(KeyError, collection.get_index, 10)

    def test_get_by_list_index(self):
        """Normal list indexing is unaffected"""
        collection = AddressCollection([
            {'input_index': 0, 'input_id': 'A'},
            {'input_index': 1, 'input_id': 'h'},
            {'input_index': 2, 'input_id': 'X'},
        ])
        self.assertEqual(collection[0].id, 'A')
        self.assertEqual(collection[1].id, 'h')
        self.assertEqual(collection[2].id, 'X')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_smartystreets
----------------------------------

Tests for `smartystreets` module.
"""

import responses
import unittest
from mock import MagicMock

from smartystreets.client import Client, validate_args, truncate_args, stringify
from smartystreets.data import Address, AddressCollection
from smartystreets.exceptions import (SmartyStreetsInputError, SmartyStreetsAuthError,
                                      SmartyStreetsPaymentError, SmartyStreetsServerError)


class TestDecorators(unittest.TestCase):
    """
    The decorator functions provide some safety around the parameters provided to API calls
    """

    def test_truncate_args(self):

        selfarg = MagicMock()
        myargs = [i for i in range(104)]

        def somefunc(myself, args):
            return len(args)

        selfarg.truncate_addresses = True
        self.assertEqual(truncate_args(somefunc)(selfarg, myargs), 100)

        selfarg.truncate_addresses = False
        modified_func = truncate_args(somefunc)
        self.assertRaises(ValueError, modified_func, selfarg, myargs)

    def test_validate_args(self):

        def somefunc(myself, args):
            return len(args)

        selfarg = MagicMock()

        myargs = ["1", "2", "3"]
        self.assertEqual(validate_args(somefunc)(selfarg, myargs), 3)

        myargs = [1, 2, 3]
        modified_func = validate_args(somefunc)
        self.assertRaises(TypeError, modified_func, selfarg, myargs)

        myargs = [{}, "?"]
        modified_func = validate_args(somefunc)
        self.assertRaises(TypeError, modified_func, selfarg, myargs)


class TestClient(unittest.TestCase):

    def setUp(self):
        self.client = Client(auth_id='blah', auth_token='blibbidy')

    @responses.activate
    def test_input_error(self):
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                  body='', status=400,
                  content_type='application/json')
        self.assertRaises(SmartyStreetsInputError, self.client.street_addresses, [{}, {}])

    @responses.activate
    def test_auth_error(self):
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='', status=401,
                      content_type='application/json')
        self.assertRaises(SmartyStreetsAuthError, self.client.street_addresses, [{}, {}])

    @responses.activate
    def test_payment_error(self):
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='', status=402,
                      content_type='application/json')
        self.assertRaises(SmartyStreetsPaymentError, self.client.street_addresses, [{}, {}])

    @responses.activate
    def test_server_error(self):
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='', status=500,
                      content_type='application/json')
        self.assertRaises(SmartyStreetsServerError, self.client.street_addresses, [{}, {}])

    @responses.activate
    def test_one_address(self):
        """Ensure singluar street address method returns an Address"""
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='[{"street_address": "100 Main St"}]',
                      status=200, content_type='application/json')
        response = self.client.street_address({"street": "100 Main st"})
        self.assertIsInstance(response, Address)

    @responses.activate
    def test_addresses_response(self):
        """Ensure address return an AddressCollection"""
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='[{"street_address": "100 Main St"}, {"street_address": "200 Main St"}]',
                      status=200, content_type='application/json')
        response = self.client.street_addresses([{"street": "100 Main st"},
                                                 {"street": "200 Main St"}])
        self.assertIsInstance(response, AddressCollection)
        self.assertEqual(2, len(response))


class TestDataValidation(unittest.TestCase):
    """
    SmartyStreets expects JSON values as strings - test that all data is updated that way
    """

    def test_stringify(self):
        self.assertEqual(
            stringify([{"input_id": 123, "zipcode": 20120, "candidates": "9"}]),
            [{"input_id": "123", "zipcode": "20120", "candidates": 9}],
        )

    def test_padded_integer(self):
        self.assertEqual(
            stringify([{"input_id": 123, "zipcode": 120, "candidates": "9"}]),
            [{"input_id": "123", "zipcode": "00120", "candidates": 9}],
        )



"""
Uses gevent to make concurrent requests.
"""

import grequests
import json
import logging

from .client import Client, validate_args, stringify
from .data import AddressCollection
from .exceptions import SmartyStreetsError, ERROR_CODES

try:
    # Python 2
    ranger = xrange
except NameError:
    # Python 3
    ranger = range


logger = logging.getLogger(__name__)


def chunker(l, n):
    """
    Generates n-sized chunks from the list l
    """
    for i in ranger(0, len(l), n):
        yield l[i:i + n]


def response_error(request, exception):
    """
    Handles a grequest connection error.

    Applies to a feature in non-packaged version of grequests.

    :param request:
    :param exception:
    """
    logger.error("Request exception: {}".format(exception))


class AsyncClient(Client):
    """
    A SmartyStreets client that supports concurrent requests.

    You are strongly recommended to use the input_id parameter as the input_index values will be
    all but worthless.
    """
    def post(self, endpoint, data, parallelism=5):
        """
        Executes most of the request.

        The parallelism parameter is useful to avoid swamping the API service with calls. Thus
        the entire set of requests won't be all made at once, but in chunked groups.

        :param endpoint: string indicating the URL component to call
        :param data: the JSON ready data to submit (list of dictionaries of addresses)
        :param parallelism: number of simultaneous requests to make.
        :return: a tuple of an AddressCollection and a dictionary of the response codes and the
                count for each.
        """
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'x-standardize-only': 'true' if self.standardize else 'false',
            'x-include-invalid': 'true' if self.invalid else 'false',
            'x-accept-keypair': 'true' if self.accept_keypair else 'false',
        }
        if not self.logging:
            headers['x-suppress-logging'] = 'false'

        params = {'auth-id': self.auth_id, 'auth-token': self.auth_token}
        url = self.BASE_URL + endpoint

        rs = (
            grequests.post(
                url=url,
                data=json.dumps(stringify(data_chunk)),
                params=params,
                headers=headers,
            ) for data_chunk in chunker(data, 100)
        )

        responses = grequests.imap(rs, size=parallelism)
        status_codes = {}
        addresses = AddressCollection([])
        for response in responses:
            if response.status_code not in status_codes.keys():
                status_codes[response.status_code] = 1
            else:
                status_codes[response.status_code] += 1

            if response.status_code == 200:
                addresses[0:0] = AddressCollection(response.json())  # Fast list insertion

            # If an auth error is raised, it's safe to say that this is
            # going to affect every request, so raise the exception immediately..
            elif response.status_code == 401:
                raise ERROR_CODES[401]

        # The return value or exception is simple if it is consistent.
        if len(status_codes.keys()) == 1:
            if 200 in status_codes:
                return addresses, status_codes
            else:
                raise ERROR_CODES.get(status_codes.keys()[0], SmartyStreetsError)

        # For any other mix not really sure of the best way to handle it. If it's a mix of 200
        # and error codes, then returning the resultant addresses and status code dictionary
        # seems pretty sensible. But if it's a mix of all error codes (could be a mix of payment
        # error, input error, potentially server error) this will probably require careful
        # checking in the code using this interface.
        return addresses, status_codes

    @validate_args
    def street_addresses(self, addresses):
        """
        Provides a consistent API across client types

        :param addresses: a list of addresses in dictionary format
        :return: a tuple of an AddressCollection and a dictionary of the response codes and the
        """
        return self.post("street-address", addresses)

    def street_address(self, address):
        raise NotImplementedError(
            "The street_address method is not implemented for the AsyncClient")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_smartystreets
----------------------------------

Tests for `smartystreets` module.

Note that for now async behavior is provided via grequests, which uses gevent, which
in packaged form is only supported on Python 2.7. As such this functionality is not
expected to work on Python 3 (for now) which means the tests here are skipped for
Python 3. This is a sad.
"""
import responses
import sys
import unittest
from mock import MagicMock

from smartystreets.data import Address, AddressCollection
from smartystreets.exceptions import (SmartyStreetsInputError, SmartyStreetsAuthError,
                                      SmartyStreetsPaymentError, SmartyStreetsServerError)

skip = False

try:
    import grequests
except ImportError:
    # We will skip these tests if we can't find grequests.  The sync. client
    # returns data in a slightly different format, so these tests fail if we
    # try to substiute the sync. for the async. client.
    skip = True
else:
    from smartystreets.async import AsyncClient


@unittest.skipIf(skip, "No grequests support found")
class TestAsyncClient(unittest.TestCase):

    def setUp(self):
        self.client = AsyncClient(auth_id='blah', auth_token='blibbidy')

    @responses.activate
    def test_input_error(self):
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='', status=400,
                      content_type='application/json')
        self.assertRaises(SmartyStreetsInputError, self.client.street_addresses, [{}, {}])

    @responses.activate
    def test_auth_error(self):
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='', status=401,
                      content_type='application/json')
        self.assertRaises(SmartyStreetsAuthError, self.client.street_addresses, [{}, {}])

    @responses.activate
    def test_payment_error(self):
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='', status=402,
                      content_type='application/json')
        self.assertRaises(SmartyStreetsPaymentError, self.client.street_addresses, [{}, {}])

    @responses.activate
    def test_server_error(self):
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='', status=500,
                      content_type='application/json')
        self.assertRaises(SmartyStreetsServerError, self.client.street_addresses, [{}, {}])

    @responses.activate
    def test_addresses_response(self):
        """Ensure address return an AddressCollection"""
        responses.add(responses.POST, 'https://api.smartystreets.com/street-address',
                      body='[{"street_address": "100 Main St"}, {"street_address": "200 Main St"}]',
                      status=200, content_type='application/json')
        response = self.client.street_addresses([{"street": "100 Main st"},
                                                 {"street": "200 Main St"}])
        self.assertIsInstance(response[0], AddressCollection)
        self.assertEqual(2, len(response[0]))
